# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import stat
import time

import paramiko
from rally.common import logging
from rally.plugins.openstack.wrappers import network as network_wrapper
from rally.task import utils as task_utils

LOG = logging.getLogger(__name__)
SUBNET_IP_VERSION = 4
START_CIDR = "10.2.0.0/24"
EXT_NET_CIDR = "172.16.1.0/24"


def execute_cmd_over_ssh(host, cmd, private_key):
    """Run the given command over ssh

    Using paramiko package, it creates a connection to the given host;
    executes the required command on it and returns the output.
    :param host: Dictionary of ip, username and password
    :param cmd: Command to be run over ssh
    :param private_key: path to private key file
    :return: Output of the executed command
    """
    LOG.debug('EXECUTE COMMAND <%s> OVER SSH', cmd)
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    k = paramiko.RSAKey.from_private_key_file(private_key)

    try:

        client.connect(host["ip"], username=host["username"], pkey=k)
    except paramiko.BadHostKeyException as e:
        raise Exception(
            "BADHOSTKEY EXCEPTION WHEN CONNECTING TO %s", host["ip"], e)
    except paramiko.AuthenticationException as e:
        raise Exception(
            "AUTHENTICATION EXCEPTION WHEN CONNECTING TO %s", host["ip"], e)
    except paramiko.SSHException as e:
        raise Exception("SSH EXCEPTION WHEN CONNECTING TO %s", host["ip"], e)
    except OSError as e:
        raise Exception("SOCKET ERROR WHEN CONNECTING TO %s", host["ip"], e)
    LOG.debug("CONNECTED TO HOST <%s>", host["ip"])
    try:
        stdin, stdout, stderr = client.exec_command(cmd)
        return stdout.read().splitlines()
    except paramiko.SSHException as e:
        raise Exception("SSHEXCEPTION WHEN CONNECTING TO %s", host["ip"], e)
    finally:
        client.close()


def create_tenant(keystone_client, tenant_suffix):
    """Creates keystone tenant with a random name.

    :param keystone_client: keystone client
    :param tenant_suffix: suffix name for the tenant
    :returns: uuid of the new tenant
    """
    tenant_name = "rally_tenant_" + tenant_suffix
    LOG.debug("CREATING NEW TENANT %s", tenant_name)
    return keystone_client.tenants.create(tenant_name).id


def create_network(neutron_client, neutron_admin_client, network_suffix,
                   tenant_id=None, DVR_flag=True, ext_net_name=None):
    """Create neutron network, subnet, router

    :param neutron_client: neutron client
    :param neutron_admin_client: neutron client with admin credentials
    :param network_suffix: str, suffix name of the new network
    :param tenant_id: uuid of the tenant
    :param DVR_flag: True - creates a DVR router
                     False - creates a non DVR router
    :param ext_net_name: external network that is to be used
    :return: router, subnet, network, subnet_cidr
    """
    subnet_cidr = network_wrapper.generate_cidr(start_cidr=START_CIDR)

    def _create_network(neutron_client, network_suffix, is_external=False):
        """Creates neutron network"""

        network_name = "rally_network_" + network_suffix
        network_args = {"name": network_name,
                        "router:external": is_external
                        }
        if tenant_id:
            network_args["tenant_id"] = tenant_id
        LOG.debug("ADDING NEW NETWORK %s", network_name)
        return neutron_client.create_network({"network": network_args})

    def _create_subnet(neutron_client, rally_network, network_suffix, cidr):
        """Create neutron subnet"""

        network_id = rally_network["network"]["id"]
        subnet_name = "rally_subnet_" + network_suffix
        subnet_args = {"name": subnet_name,
                       "cidr": cidr,
                       "network_id": network_id,
                       "ip_version": SUBNET_IP_VERSION
                       }
        if tenant_id:
            subnet_args["tenant_id"] = tenant_id
        LOG.debug("ADDING SUBNET %s", subnet_name)
        return neutron_client.create_subnet({"subnet": subnet_args})

    def _create_router(neutron_client, ext_network_id, rally_subnet, dvr_flag):
        """Create router, set the external gateway and add router interface

        :param neutron_client: neutron_client
        :param ext_network_id: uuid of the external network
        :param rally_subnet: subnet to add router interface
        :param dvr_flag: True - creates a DVR router
                         False - creates a non DVR router
        :return: router
        """
        router_name = "rally_router_" + network_suffix
        gw_info = {"network_id": ext_network_id}
        router_args = {"name": router_name,
                       "external_gateway_info": gw_info
                       }
        if not dvr_flag:
            router_args["distributed"] = dvr_flag
        if tenant_id:
            router_args["tenant_id"] = 'tenant_id'
        LOG.debug("ADDING ROUTER %s", router_name)
        rally_router = neutron_client.create_router({"router": router_args})

        LOG.debug("[%s]: ADDING ROUTER INTERFACE")
        neutron_client.add_interface_router(
            rally_router['router']["id"],
            {"subnet_id": rally_subnet["subnet"]["id"]})
        return rally_router

    def _get_external_network_id(ext_net_name):
        """Fetch the network id for the given external network, if it exists.
           Else fetch the first external network present.
        """

        ext_nets = neutron_client.list_networks(
            **{'router:external': True})['networks']

        ext_nets_searched = [n for n in ext_nets if n['name'] == ext_net_name]
        if ext_nets_searched:
            return ext_nets_searched[0]['id']
        elif ext_nets:
            return ext_nets[0]['id']
        else:
            return None

    def _create_external_network():
        """Creat external network and subnet"""

        ext_net = _create_network(neutron_admin_client, "public", True)
        _create_subnet(neutron_admin_client, ext_net, "public", EXT_NET_CIDR)
        return ext_net['network']['id']

    ext_network_id = _get_external_network_id(ext_net_name)
    if not ext_network_id:
        ext_network_id = _create_external_network()
    rally_network = _create_network(neutron_client, network_suffix)
    rally_subnet = _create_subnet(neutron_client, rally_network,
                                  network_suffix, subnet_cidr)
    rally_router = _create_router(neutron_client, ext_network_id,
                                  rally_subnet, DVR_flag)
    return rally_router, rally_network, rally_subnet, subnet_cidr


def create_keypair(nova_client, keypair_suffix):
    """Create keypair

    :param nova_client: nova_client
    :param keypair_suffix: sufix name for the keypair
    :return: keypair
    """
    keypair_name = "rally_keypair_" + keypair_suffix
    LOG.debug("CREATING A KEYPAIR %s", keypair_name)
    keypair = nova_client.keypairs.create(keypair_name)
    return keypair


def write_key_to_local_path(keypair, local_key_file):
    """Write the private key of the nova instance to a temp file

    :param keypair: nova keypair
    :param local_key_file: path to private key file
    :return:
    """

    with open(local_key_file, 'w') as f:
        os.chmod(local_key_file, stat.S_IREAD | stat.S_IWRITE)
        f.write(keypair.private_key)


def write_key_to_compute_node(keypair, local_path, remote_path, host,
                              private_key):
    """Write the private key of the nova instance to the compute node

    First fetches the private key from the keypair and writes it to a
    temporary file in the local machine. It then sftp's the file
    to the compute host.

    :param keypair: nova keypair
    :param local_path: path to private key file of the nova instance in the
                       local machine
    :param remote_path: path where the private key file has to be placed
                        in the remote machine
    :param host: compute host credentials
    :param private_key: path to your private key file
    :return:
    """

    LOG.debug("WRITING PRIVATE KEY TO COMPUTE NODE")
    k = paramiko.RSAKey.from_private_key_file(private_key)
    write_key_to_local_path(keypair, local_path)
    try:
        transport = paramiko.Transport(host['ip'], host['port'])
    except paramiko.SSHException as e:
        raise Exception(
            "PARAMIKO TRANSPORT FAILED. CHECK IF THE HOST IP %s AND PORT %s "
            "ARE CORRECT %s", host['ip'], host['port'], e)
    try:
        transport.connect(
                username=host['username'], pkey=k)
    except paramiko.BadHostKeyException as e:
        transport.close()
        raise Exception(
            "BADHOSTKEY EXCEPTION WHEN CONNECTING TO %s", host["ip"], e)
    except paramiko.AuthenticationException as e:
        transport.close()
        raise Exception("AUTHENTICATION EXCEPTION WHEN CONNECTING TO %s",
                        host["ip"], e)
    except paramiko.SSHException as e:
        transport.close()
        raise Exception("SSH EXCEPTION WHEN CONNECTING TO %s", host["ip"], e)
    LOG.debug("CONNECTED TO HOST <%s>", host["ip"])

    try:
        sftp_client = paramiko.SFTPClient.from_transport(transport)
        sftp_client.put(local_path, remote_path)
    except OSError as e:
        raise Exception("FILE PATH DOESN'T EXIST", e)
    finally:
        transport.close()


def create_server(nova_client, keypair, **kwargs):
    """Create nova instance

    :param nova_client: nova client
    :param keypair: key-pair to allow ssh
    :return: new nova instance
    """
    # add sec-group
    sec_group_name = "rally_secgroup_" + kwargs["sec_group_suffix"]
    LOG.debug("ADDING NEW SECURITY GROUP %s", sec_group_name)
    secgroup = nova_client.security_groups.create(sec_group_name,
                                                  sec_group_name)
    # add security rules for SSH and ICMP
    nova_client.security_group_rules.create(secgroup.id, from_port=22,
                to_port=22, ip_protocol="tcp", cidr="0.0.0.0/0")

    nova_client.security_group_rules.create(secgroup.id, from_port=-1,
                to_port=-1, ip_protocol="icmp", cidr="0.0.0.0/0")

    # boot new nova instance
    server_name = "rally_server_" + (kwargs["server_suffix"])
    LOG.debug("BOOTING NEW INSTANCE: %s", server_name)
    LOG.debug("%s", kwargs["image"])
    server = nova_client.servers.create(server_name,
                                        image=kwargs["image"],
                                        flavor=kwargs["flavor"],
                                        key_name=keypair.name,
                                        security_groups=[secgroup.id],
                                        nics=kwargs["nics"])
    return server


def assert_server_status(server, **kwargs):
    """Assert server status

    :param server: nova server
    """

    LOG.debug('WAITING FOR SERVER TO GO ACTIVE')
    server = task_utils.wait_for(
        server,
        is_ready=task_utils.resource_is("ACTIVE"),
        update_resource=task_utils.get_from_manager(),
        timeout=kwargs["nova_server_boot_timeout"],
        check_interval=5)
    LOG.debug("SERVER STATUS: %s", server.status)
    assert ('ACTIVE' == server.status), ("THE INSTANCE IS NOT IN ACTIVE STATE")


def get_server_ip(nova_client, server_id, network_suffix):
    """Get the ip associated with the nova instance

    :param nova_client: nova client
    :param server_id: uuid of the nova instance whose ip is required
    :param network_suffix: suffix name of the network
    :return: ip address of the instance
    """
    network_name = "rally_network_" + network_suffix
    server_details = nova_client.servers.get(server_id)
    server_ip = server_details.addresses[network_name][0]["addr"]
    return server_ip


def add_floating_ip(nova_client, server):
    """Associates floating-ip to a server

    :param nova_client: nova client
    :param server: nova instance
    :return: associated floating ip
    """

    fip_list = nova_client.floating_ips.list()
    for fip in fip_list:
        if fip.instance_id is None:
            floating_ip = fip
            break
    else:
        LOG.debug("CREATING NEW FLOATING IP")
        floating_ip = nova_client.floating_ips.create()
    LOG.debug("ASSOCIATING FLOATING IP %s", floating_ip.ip)
    nova_client.servers.add_floating_ip(server.id, floating_ip.ip)
    return floating_ip


def get_namespace(host, private_key):
    """SSH into the host and get the namespaces

    :param host : dictionary of controller/compute node credentials
     {ip:x.x.x.x, username:xxx, password:xxx}
    :param private_key: path to private key file
    :return: namespaces
    """
    LOG.debug("GET NAMESPACES")
    cmd = "sudo ip netns"
    namespaces = execute_cmd_over_ssh(host, cmd, private_key)
    LOG.debug("NAMESPACES %s", namespaces)
    return namespaces


def wait_for_namespace_creation(namespace_tag, router_id, hosts, private_key,
                                timeout=60):
    """Wait for the namespace creation

    Get into each of the controllers/compute nodes and check which one contains
    the snat/qrouter namespace corresponding to rally_router. Sleep for a sec
    and repeat until either the namespace is found or the namespace_creation_
    time exceeded.
    :param namespace_tag: which namespace ("snat_" or "qrouter_")
    :param router_id: uuid of the rally_router
    :param hosts: controllers or compute hosts
    :param private_key: path to private key file
    :param timeout: namespace creation time
    :return:
    """
    start_time = time.time()
    while True:
        for host in hosts:
            namespaces = get_namespace(host, private_key)
            for line in namespaces:
                if line == (namespace_tag + router_id):
                    namespace_tag = line
                    return namespace_tag, host
        time.sleep(1)
        if time.time() - start_time > timeout:
            raise Exception("TIMEOUT WHILE WAITING FOR"
                            " NAMESPACES TO BE CREATED")


def ping(host, cmd, private_key):
    """Execute ping command over ssh"""
    ping_result = execute_cmd_over_ssh(host, cmd, private_key)
    if ping_result:
        LOG.debug("PING RESULT %s", ping_result)
        return True
    else:
        return False


def ping_router_gateway(namespace_controller_tuple, router_gw_ip, private_key):
    """Ping the ip address from network namespace

    Get into controller's snat-namespaces and ping the peer router gateway ip.
    :param namespace_controller_tuple: namespace, controller tuple. (It's the
                                 controller that contains the namespace )
    :param router_gw_ip: ip address to be pinged
    :param private_key: path to private key file
    :return: True if ping succeeds
             False if ping fails
    """
    namespace, controller = namespace_controller_tuple
    LOG.debug("PING %s FROM THE NAMESPACE %s", router_gw_ip, namespace)
    count = 4
    cmd = "sudo ip netns exec {} ping -w {} -c {} {}".format(
        namespace, 2 * count, count, router_gw_ip)
    return ping(controller, cmd, private_key)


def get_interfaces(namespace_controller_tuple, private_key):
    """Get the interfaces

    Get into the controller's snat namespace and list the interfaces.
    :param namespace_controller_tuple: namespace, controller tuple(the
                                       controller that contains the namespace).
    :param private_key: path to private key file
    :return: interfaces
    """
    namespace, controller = namespace_controller_tuple
    LOG.debug("GET THE INTERFACES BY USING 'ip a' FROM THE NAMESPACE %s",
              namespace)
    cmd = f"sudo ip netns exec {namespace} ip a"
    interfaces = execute_cmd_over_ssh(controller, cmd, private_key)
    LOG.debug("INTERFACES %s", interfaces)
    return interfaces


def start_tcpdump(namespace_controller_tuple, interface, private_key):
    """Start the tcpdump at the given interface

    Get into the controller's snat namespace and start a tcp dump at the
    qg-interface.
    :param namespace_controller_tuple: namespace, controller tuple. (It's the
                                 controller that contains the namespace )
    :param interface: interface in which tcpdump has to be run
    :param private_key: path to private key file
    :return: tcpdump output
    """
    namespace, controller = namespace_controller_tuple
    LOG.debug("START THE TCPDUMP USING 'tcpdump -i %s FROM THE NAMESPACE"
              " %s", interface, namespace)
    cmd = ("sudo ip netns exec {} timeout 15 tcpdump -n -i {}"
           .format(namespace, interface))
    tcpdump = execute_cmd_over_ssh(controller, cmd, private_key)
    LOG.debug("TCPDUMP %s", tcpdump)
    return tcpdump


def ssh_and_ping_server(local_server, peer_server, ns_compute_tuple, keyfile,
                        private_key):
    """SSH and ping the nova instance from the namespace

     Get into the compute node's qrouter namespace and then ssh into the local
     nova instance & ping the peer nova instance.
    :param local_server: private ip of the server to ssh into
    :param peer_server: private ip of the server to ping to
    :param ns_compute_tuple: namespace, compute tuple. (It's the
                                 compute node that contains the namespace )
    :param keyfile: path to private key file of the nova instance
    :param private_key: path to private key file
    :return: True if ping succeeds
             False if ping fails
    """
    namespace, compute_host = ns_compute_tuple
    LOG.debug("SSH INTO SERVER %s AND PING THE PEER SERVER %s FROM THE"
              " NAMESPACE %s", local_server, peer_server, namespace)
    host = "cirros@" + local_server
    count = 20
    cmd = ("sudo ip netns exec {} ssh -v -o StrictHostKeyChecking=no -o"
           "HashKnownHosts=no -i {} {} ping -w {} -c {} {}"
           .format(namespace, keyfile, host, 2 * count, count, peer_server))
    return ping(compute_host, cmd, private_key)


def ssh_and_ping_server_with_fip(local_server, peer_server, keyfile,
                                 private_key):
    """SSH into the local nova instance and ping the peer instance using fips

    :param local_server: fip of the server to ssh into
    :param peer_server: private ip of the server to ping to
    :param keyfile: path to private key file of the nova instance
    :param private_key: path to private key file
    :return: True if ping succeeds
             False if ping fails
    """
    LOG.debug("SSH INTO LOCAL SERVER %s AND PING THE PEER SERVER %s",
              local_server.ip, peer_server)
    count = 20
    local_host = {"ip": "127.0.0.1", "username": None}
    host = "cirros@" + local_server.ip
    cmd = ("ssh -v -o StrictHostKeyChecking=no -o"
           "HashKnownHosts=no -i {} {} ping -w {} -c {} {}"
           .format(keyfile, host, 2 * count, count, peer_server))
    return ping(local_host, cmd, private_key)


def delete_servers(nova_client, servers):
    """Delete nova servers

    It deletes the nova servers, associated security groups.

    :param nova_client: nova client
    :param servers: nova instances to be deleted
    :return:
    """
    for server in servers:
        LOG.debug("DELETING NOVA INSTANCE: %s", server.id)
        sec_group_id = server.security_groups[0]['name']
        nova_client.servers.delete(server.id)

        LOG.debug("WAITING FOR INSTANCE TO GET DELETED")
        task_utils.wait_for_delete(
            server, update_resource=task_utils.get_from_manager())

        for secgroup in nova_client.security_groups.list():
            if secgroup.id == sec_group_id:
                LOG.debug("DELETING SEC_GROUP: %s", sec_group_id)
                nova_client.security_groups.delete(secgroup.id)


def delete_floating_ips(nova_client, fips):
    """Delete floating ips

    :param nova_client: nova client
    :param fips: list of floating ips
    :return:
    """
    for fip in fips:
        nova_client.floating_ips.delete(fip.id)


def delete_keypairs(nova_client, keypairs):
    """Delete key pairs

    :param nova_client: nova client
    :param keypairs: list of keypairs
    :return
    """
    for key_pair in keypairs:
        LOG.debug("DELETING KEY_PAIR %s", key_pair.name)
        nova_client.keypairs.delete(key_pair.id)


def delete_networks(neutron_client, neutron_admin_client,
                   routers, networks, subnets):
    """Delete neutron network, subnets amd routers

    :param neutron_client: neutron client
    :param neutron_admin_client: neutron_admin_client
    :param routers: list of routers to be deleted
    :param networks: list of networks to be deleted
    :param subnets: list of subnets to be deleted
    :return
    """
    LOG.debug("DELETING RALLY ROUTER INTERFACES & GATEWAYS")
    for router in routers:
        neutron_client.remove_gateway_router(router['router']['id'])
        router_name = router['router']['name']
        subnet_name = ("rally_subnet_" + router_name[13:len(router_name)])
        for subnet in subnets:
            if subnet_name == subnet['subnet']['name']:
                neutron_client.remove_interface_router(
                    router['router']['id'],
                    {"subnet_id": subnet['subnet']['id']})

    LOG.debug("DELETING RALLY ROUTERS")
    for router in routers:
        neutron_client.delete_router(router['router']['id'])

    LOG.debug("DELETING RALLY NETWORKS")
    for network in networks:
        if (network['network']['router:external'] and
            network['network']['name'] == "rally_network_public"):
            external_network = network
            neutron_admin_client.delete_network(
                external_network['network']["id"])
        elif network['network']['router:external']:
            pass
        else:
            neutron_client.delete_network(network['network']['id'])


def delete_tenants(keystone_client, tenant_ids):
    """Delete keystone tenant

    :param keystone_client: keystone client
    :param tenant_ids: list of tenants' uuids
    :returns: delete keystone tenant instance
    """
    LOG.debug('DELETE TENANTS')
    for id in tenant_ids:
        keystone_client.tenants.delete(id)


def delete_keyfiles(local_key_files, remote_key_files=None,
                    ns_compute_tuples=None, private_key=None):
    """Delete the SSH keyfiles from the compute and the local nodes

    :param local_key_files: paths to ssh key files in local node
    :param remote_key_files: paths to ssh key files in compute nodes
    :param ns_compute_tuples: namespace, compute tuple. (It's the
                              compute node that contains the namespace )
    :param private_key: path to private key file
    :return:
    """
    LOG.debug("DELETING RALLY KEY FILES FROM LOCAL MACHINE")
    for key in local_key_files:
        if os.path.exists(key):
            os.remove(key)

    if ns_compute_tuples:
        LOG.debug("DELETING RALLY KEY FILES FROM COMPUTE HOSTS")
        for key, ns_comp in zip(remote_key_files, ns_compute_tuples):
            cmd = f"sudo rm -f {key}"
            host = ns_comp[1]
            execute_cmd_over_ssh(host, cmd, private_key)


def delete_hosts_from_knownhosts_file(hosts, ns_compute_tuples=None,
                                      private_key=None):
    """Remove the hosts from the knownhosts file

    :param hosts: host ips to be removed from /root/.ssh/knownhosts
    :param ns_compute_tuples: namespace, compute tuple. (It's the
                              compute node that contains the namespace )
    :param private_key: path to private key file
    :return:
    """
    if ns_compute_tuples:
        LOG.debug("DELETES HOSTS FROM THE KNOWNHOSTS FILE")
        for host, ns_comp in zip(hosts, ns_compute_tuples):
            compute_host = ns_comp[1]
            cmd = ("sudo ssh-keygen -f /root/.ssh/known_hosts -R"
                   " {}".format(host))
            execute_cmd_over_ssh(compute_host, cmd, private_key)
    else:
        for host in hosts:
            os.system("sudo ssh-keygen -f /root/.ssh/known_hosts -R"
                      " {}".format(host))



# Copyright(c) 2015, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""fix identifier map fk

Revision ID: 56893333aa52
Revises: kilo
Create Date: 2015-06-11 12:09:01.263253

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import column
from sqlalchemy.sql import expression as expr
from sqlalchemy.sql import func
from sqlalchemy.sql import table

from neutron.db import migration
from neutron.db.migration import cli


# revision identifiers, used by Alembic.
revision = '56893333aa52'
down_revision = 'kilo'
branch_labels = (cli.CONTRACT_BRANCH,)


def upgrade():
    insp = sa.inspect(op.get_bind())
    if 'cisco_csr_identifier_map' not in insp.get_table_names():
        return
    # re-size existing data if necessary
    identifier_map = table('cisco_csr_identifier_map',
                           column('ipsec_site_conn_id', sa.String(36)))
    ipsec_site_conn_id = identifier_map.columns['ipsec_site_conn_id']

    op.execute(identifier_map.update(values={
        ipsec_site_conn_id: expr.case([(func.length(ipsec_site_conn_id) > 36,
                                      func.substr(ipsec_site_conn_id, 1, 36))],
                                      else_=ipsec_site_conn_id)}))

    # Need to drop foreign key constraint before mysql will allow changes

    with migration.remove_fks_from_table('cisco_csr_identifier_map'):
        op.alter_column(table_name='cisco_csr_identifier_map',
                        column_name='ipsec_site_conn_id',
                        type_=sa.String(36),
                        existing_nullable=False)



# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc
import base64
import copy
import filecmp
import os
import re
import shutil
import socket
import sys
import typing as ty

import eventlet
import jinja2
import netaddr
from neutron.agent.l3.router_info import RouterInfo
from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils as agent_utils
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.exceptions import vpn as vpn_exception
from neutron_lib.plugins import utils as plugin_utils
from neutron_lib import rpc as n_rpc
from neutron_lib.utils import file as file_utils
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
import oslo_messaging
from oslo_service import loopingcall
from oslo_utils import encodeutils
from oslo_utils import fileutils

from neutron_vpnaas._i18n import _
from neutron_vpnaas.services.vpn.common import topics
from neutron_vpnaas.services.vpn import device_drivers

LOG = logging.getLogger(__name__)
TEMPLATE_PATH = os.path.dirname(os.path.abspath(__file__))

ipsec_opts = [
    cfg.StrOpt(
        'config_base_dir',
        default='$state_path/ipsec',
        help=_('Location to store ipsec server config files')),
    cfg.IntOpt('ipsec_status_check_interval',
               default=60,
               help=_("Interval for checking ipsec status")),
    cfg.BoolOpt('enable_detailed_logging',
                default=False,
                help=_("Enable detail logging for ipsec pluto process. "
                       "If the flag set to True, the detailed logging will "
                       "be written into config_base_dir/<pid>/log. "
                       "Note: This setting applies to OpenSwan and LibreSwan "
                       "only. StrongSwan logs to syslog.")),
]
cfg.CONF.register_opts(ipsec_opts, 'ipsec')

openswan_opts = [
    cfg.StrOpt(
        'ipsec_config_template',
        default=os.path.join(
            TEMPLATE_PATH,
            'template/openswan/ipsec.conf.template'),
        help=_('Template file for ipsec configuration')),
    cfg.StrOpt(
        'ipsec_secret_template',
        default=os.path.join(
            TEMPLATE_PATH,
            'template/openswan/ipsec.secret.template'),
        help=_('Template file for ipsec secret configuration'))
]

cfg.CONF.register_opts(openswan_opts, 'openswan')

pluto_opts = [
    cfg.IntOpt('shutdown_check_timeout',
               default=1,
               help=_('Initial interval in seconds for checking if pluto '
                      'daemon is shutdown'),
               deprecated_group='libreswan'),
    cfg.IntOpt('shutdown_check_retries',
               default=5,
               help=_('The maximum number of retries for checking for '
                      'pluto daemon shutdown'),
               deprecated_group='libreswan'),
    cfg.FloatOpt('shutdown_check_back_off',
                 default=1.5,
                 help=_('A factor to increase the retry interval for '
                        'each retry'),
                 deprecated_group='libreswan'),
    cfg.BoolOpt('restart_check_config',
                default=False,
                help=_('Enable this flag to avoid from unnecessary restart'),
                deprecated_group='libreswan')
]

cfg.CONF.register_opts(pluto_opts, 'pluto')

JINJA_ENV = None

IPSEC_CONNS = 'ipsec_site_connections'

# *Swan supports Base64 encoded binary values as PSKs. In such cases,
# a character sequence beginning with 0s is interpreted as Base64
# encoded binary data.
#   - StrongSwan
#     - https://wiki.strongswan.org/projects/strongswan/wiki/PskSecret
#   - LibreSwan
#     - https://libreswan.org/man/ipsec.secrets.5.html
#     - https://libreswan.org/man/ipsec_ttodata.3.html
#   - OpenSwan (no online documents, see manpages sources in the repository)
#     - https://github.com/xelerance/Openswan
PSK_BASE64_PREFIX = '0s'


def _get_template(template_file):
    global JINJA_ENV
    if not JINJA_ENV:
        templateLoader = jinja2.FileSystemLoader(searchpath="/")
        JINJA_ENV = jinja2.Environment(loader=templateLoader, autoescape=True)
    return JINJA_ENV.get_template(template_file)


class BaseSwanProcess(metaclass=abc.ABCMeta):
    """Swan Family Process Manager

    This class manages start/restart/stop ipsec process.
    This class create/delete config template
    """

    binary = "ipsec"
    CONFIG_DIRS = [
        'var/run',
        'log',
        'etc',
        'etc/ipsec.d/aacerts',
        'etc/ipsec.d/acerts',
        'etc/ipsec.d/cacerts',
        'etc/ipsec.d/certs',
        'etc/ipsec.d/crls',
        'etc/ipsec.d/ocspcerts',
        'etc/ipsec.d/policies',
        'etc/ipsec.d/private',
        'etc/ipsec.d/reqs',
        'etc/pki/nssdb/'
    ]

    DIALECT_MAP = {
        "3des": "3des",
        "aes-128": "aes128",
        "aes-192": "aes192",
        "aes-256": "aes256",
        "aes-ctr-128": "aesctr128",
        "aes-ctr-192": "aesctr192",
        "aes-ctr-256": "aesctr256",
        "aes-xcbc": "aes_xcbc",
        "sha256": "sha2_256",
        "sha384": "sha2_384",
        "sha512": "sha2_512",
        "group2": "modp1024",
        "group5": "modp1536",
        "group14": "modp2048",
        "group15": "modp3072",
        "group16": "modp4096",
        "group17": "modp6144",
        "group18": "modp8192",
        "group19": "ecp256",
        "group20": "ecp384",
        "group21": "ecp521",
        "group22": "dh22",
        "group23": "dh23",
        "group24": "dh24",
        "group31": "curve25519",
        "bi-directional": "start",
        "response-only": "add",
        "v2": "insist",
        "v1": "never"
    }

    NS_WRAPPER = 'neutron-vpn-netns-wrapper'

    STATUS_DICT = {
        'erouted': constants.ACTIVE,
        'unrouted': constants.DOWN
    }
    STATUS_RE = r'\d\d\d "([a-f0-9\-]+).* (unrouted|erouted);'
    STATUS_NOT_RUNNING_RE = 'Command:.*ipsec.*status.*Exit code: [1|3]$'
    STATUS_IPSEC_SA_ESTABLISHED_RE = (
        r'\d{3} #\d+: "([a-f0-9\-]+).*established.*newest')
    STATUS_IPSEC_SA_ESTABLISHED_RE2 = (
        r'\d{3} #\d+: "([a-f0-9\-\/x]+).*established.*newest')

    def __init__(self, conf, process_id, vpnservice, namespace):
        self.conf = conf
        self.id = process_id
        self.updated_pending_status = False
        self.namespace = namespace
        self.connection_status = {}
        self.config_dir = os.path.join(
            self.conf.ipsec.config_base_dir, self.id)
        self.etc_dir = os.path.join(self.config_dir, 'etc')
        self.log_dir = os.path.join(self.config_dir, 'log')
        self.update_vpnservice(vpnservice)
        self.STATUS_PATTERN = re.compile(self.STATUS_RE)
        self.STATUS_NOT_RUNNING_PATTERN = re.compile(
            self.STATUS_NOT_RUNNING_RE)
        self.STATUS_IPSEC_SA_ESTABLISHED_PATTERN = re.compile(
            self.STATUS_IPSEC_SA_ESTABLISHED_RE)
        self.STATUS_IPSEC_SA_ESTABLISHED_PATTERN2 = re.compile(
            self.STATUS_IPSEC_SA_ESTABLISHED_RE2)
        self.STATUS_MAP = self.STATUS_DICT

    def translate_dialect(self):
        if not self.vpnservice:
            return
        for ipsec_site_conn in self.vpnservice['ipsec_site_connections']:
            self._dialect(ipsec_site_conn, 'initiator')
            self._dialect(ipsec_site_conn['ikepolicy'], 'ike_version')
            for key in ['encryption_algorithm',
                        'auth_algorithm',
                        'pfs']:
                self._dialect(ipsec_site_conn['ikepolicy'], key)
                self._dialect(ipsec_site_conn['ipsecpolicy'], key)
            if (('local_id' not in ipsec_site_conn.keys()) or
                (not ipsec_site_conn['local_id'])):
                ipsec_site_conn['local_id'] = ipsec_site_conn['external_ip']

    def base64_encode_psk(self):
        if not self.vpnservice:
            return
        for ipsec_site_conn in self.vpnservice['ipsec_site_connections']:
            psk = ipsec_site_conn['psk']
            encoded_psk = base64.b64encode(encodeutils.safe_encode(psk))
            # NOTE(huntxu): base64.b64encode returns an instance of 'bytes'
            # in Python 3, convert it to a str. For Python 2, after calling
            # safe_decode, psk is converted into a unicode not containing any
            # non-ASCII characters so it doesn't matter.
            psk = encodeutils.safe_decode(encoded_psk, incoming='utf_8')
            ipsec_site_conn['psk'] = PSK_BASE64_PREFIX + psk

    def get_ns_wrapper(self):
        """
        Check if we're inside a virtualenv. If we are, then we should
        respect this and launch wrapper from venv as well.
        """
        if (hasattr(sys, 'real_prefix') or
            (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix)):
            ns_wrapper = os.path.join(sys.prefix, "bin/", self.NS_WRAPPER)
        else:
            ns_wrapper = self.NS_WRAPPER
        return ns_wrapper

    def update_vpnservice(self, vpnservice):
        self.vpnservice = vpnservice
        self.translate_dialect()
        self.base64_encode_psk()

    def _dialect(self, obj, key):
        obj[key] = self.DIALECT_MAP.get(obj[key], obj[key])

    @abc.abstractmethod
    def ensure_configs(self):
        pass

    def ensure_config_file(self, kind, template, vpnservice, file_mode=None):
        """Update config file,  based on current settings for service."""
        config_str = self._gen_config_content(template, vpnservice)
        config_file_name = self._get_config_filename(kind)
        if file_mode is None:
            file_utils.replace_file(config_file_name, config_str)
        else:
            file_utils.replace_file(config_file_name, config_str, file_mode)

    def remove_config(self):
        """Remove whole config file."""
        agent_utils.execute(
            cmd=["rm", "-rf", self.config_dir], run_as_root=True)

    def _get_config_filename(self, kind):
        config_dir = self.etc_dir
        return os.path.join(config_dir, kind)

    def ensure_config_dir(self, vpnservice):
        """Create config directory if it does not exist."""
        fileutils.ensure_tree(self.config_dir, 0o755)
        for subdir in self.CONFIG_DIRS:
            dir_path = os.path.join(self.config_dir, subdir)
            fileutils.ensure_tree(dir_path, 0o755)

    def _gen_config_content(self, template_file, vpnservice):
        template = _get_template(template_file)
        return template.render(
            {'vpnservice': vpnservice,
             'state_path': self.conf.state_path})

    def _get_rootwrap_config(self):
        if 'neutron-rootwrap' in cfg.CONF.AGENT.root_helper:
            rh_tokens = cfg.CONF.AGENT.root_helper.split(' ')
            if len(rh_tokens) == 3 and os.path.exists(rh_tokens[2]):
                return rh_tokens[2]
        return None

    @abc.abstractmethod
    def get_status(self):
        pass

    @property
    def status(self):
        if self.active:
            return constants.ACTIVE
        return constants.DOWN

    @property
    def active(self):
        """Check if the process is active or not."""
        if not self.namespace:
            return False
        try:
            status = self.get_status()
            self._extract_and_record_connection_status(status)
            if not self.connection_status:
                return False
        except RuntimeError:
            return False
        return True

    def update(self):
        """Update Status based on vpnservice configuration."""

        # Disable the process if a vpnservice is disabled or it has no
        # enabled IPSec site connections.
        vpnservice_has_active_ipsec_site_conns = any(
            ipsec_site_conn['admin_state_up']
            for ipsec_site_conn in self.vpnservice['ipsec_site_connections'])
        if (not self.vpnservice['admin_state_up'] or
                not vpnservice_has_active_ipsec_site_conns):
            self.disable()
        else:
            self.enable()

        if plugin_utils.in_pending_status(self.vpnservice['status']):
            self.updated_pending_status = True

        self.vpnservice['status'] = self.status
        for ipsec_site_conn in self.vpnservice['ipsec_site_connections']:
            if plugin_utils.in_pending_status(ipsec_site_conn['status']):
                conn_id = ipsec_site_conn['id']
                conn_status = self.connection_status.get(conn_id)
                if not conn_status:
                    continue
                conn_status['updated_pending_status'] = True
                ipsec_site_conn['status'] = conn_status['status']

    def enable(self):
        """Enabling the process."""
        try:
            self.ensure_configs()
            if self.active:
                self.restart()
            else:
                self.start()
        except RuntimeError:
            LOG.exception(
                "Failed to enable vpn process on router %s",
                self.id)

    def disable(self):
        """Disabling the process."""
        try:
            if self.active:
                self.stop()
            self.remove_config()
        except RuntimeError:
            LOG.exception(
                "Failed to disable vpn process on router %s",
                self.id)

    @abc.abstractmethod
    def restart(self):
        """Restart process."""

    @abc.abstractmethod
    def start(self):
        """Start process."""

    @abc.abstractmethod
    def stop(self):
        """Stop process."""

    def _check_status_line(self, line):
        """Parse a line and search for status information.

        If a connection has an established Security Association,
        it will be considered ACTIVE. Otherwise, even if a status
        line shows that a connection is active, it will be marked
        as DOWN-ed.
        """

        # pluto is not running so just exit
        if self.STATUS_NOT_RUNNING_PATTERN.search(line):
            self.connection_status = {}
            raise StopIteration()

        m = self.STATUS_IPSEC_SA_ESTABLISHED_PATTERN.search(line)
        if m:
            connection_id = m.group(1)
            return connection_id, constants.ACTIVE
        else:
            m = self.STATUS_PATTERN.search(line)
            if m:
                connection_id = m.group(1)
                return connection_id, constants.DOWN
        return None, None

    def _extract_and_record_connection_status(self, status_output):
        if not status_output:
            self.connection_status = {}
            return
        for line in status_output.split('\n'):
            try:
                conn_id, conn_status = self._check_status_line(line)
            except StopIteration:
                break
            if conn_id:
                self._record_connection_status(conn_id, conn_status)

    def _record_connection_status(self, connection_id, status,
                                  force_status_update=False):
        conn_info = self.connection_status.get(connection_id)
        if not conn_info:
            self.connection_status[connection_id] = {
                'status': status,
                'updated_pending_status': force_status_update
            }
        else:
            conn_info['status'] = status
            if force_status_update:
                conn_info['updated_pending_status'] = True


class OpenSwanProcess(BaseSwanProcess):
    """OpenSwan Process manager class.

    This process class uses three commands
    (1) ipsec pluto:  IPsec IKE keying daemon
    (2) ipsec addconn: Adds new ipsec addconn
    (3) ipsec whack:  control interface for IPSEC keying daemon
    """
    # Derived classes can set IPSEC_CONF_NAT_TRAVERSAL to None to
    # omit the nat_traversal line in ipsec.conf
    IPSEC_CONF_NAT_TRAVERSAL = "yes"

    def __init__(self, conf, process_id, vpnservice, namespace):
        super().__init__(conf, process_id, vpnservice, namespace)
        self.secrets_file = os.path.join(
            self.etc_dir, 'ipsec.secrets')
        self.config_file = os.path.join(
            self.etc_dir, 'ipsec.conf')
        self.pid_path = os.path.join(
            self.config_dir, 'var', 'run', 'pluto')
        self.pid_file = '%s.pid' % self.pid_path

    def _execute(self, cmd, check_exit_code=True, extra_ok_codes=None):
        """Execute command on namespace."""
        ip_wrapper = ip_lib.IPWrapper(namespace=self.namespace)
        return ip_wrapper.netns.execute(cmd, check_exit_code=check_exit_code,
                                        extra_ok_codes=extra_ok_codes)

    def ensure_configs(self):
        """Generate config files which are needed for OpenSwan.

        If there is no directory, this function will create
        dirs.
        """
        self.ensure_config_dir(self.vpnservice)
        self.ensure_config_file(
            'ipsec.conf',
            self.conf.openswan.ipsec_config_template,
            self.vpnservice)
        self.ensure_config_file(
            'ipsec.secrets',
            self.conf.openswan.ipsec_secret_template,
            self.vpnservice,
            0o600)

    def _copy_configs(self):
        if not cfg.CONF.pluto.restart_check_config:
            return
        config_file_name = self._get_config_filename('ipsec.conf')
        if os.path.isfile(config_file_name):
            shutil.copyfile(config_file_name, config_file_name + '.old')
        config_file_name = self._get_config_filename('ipsec.secrets')
        if os.path.isfile(config_file_name):
            shutil.copyfile(config_file_name, config_file_name + '.old')
        os.chmod(config_file_name + '.old', 0o600)

    def _process_running(self):
        """Checks if process is still running."""

        # If no PID file, we assume the process is not running.
        if not os.path.exists(self.pid_file):
            return False

        try:
            # We take an ask-forgiveness-not-permission approach and rely
            # on throwing to tell us something. If the pid file exists,
            # delve into the process information and check if it matches
            # our expected command line.
            with open(self.pid_file) as f:
                pid = f.readline().strip()
                with open('/proc/%s/cmdline' % pid) as cmd_line_file:
                    cmd_line = cmd_line_file.readline()
                    if self.pid_path in cmd_line and 'pluto' in cmd_line:
                        # Okay the process is probably a pluto process
                        # and it contains the pid_path in the command
                        # line... could be a race. Log to error and return
                        # that it is *NOT* okay to clean up files. We are
                        # logging to error instead of debug because it
                        # indicates something bad has happened and this is
                        # valuable information for figuring it out.
                        LOG.error('Process %(pid)s exists with command '
                                  'line %(cmd_line)s.',
                                  {'pid': pid, 'cmd_line': cmd_line})
                        return True

        except OSError as e:
            # This is logged as "info" instead of error because it simply
            # means that we couldn't find the files to check on them.
            LOG.info('Unable to find control files on startup for '
                     'router %(router)s: %(msg)s',
                     {'router': self.id, 'msg': e})
        return False

    def _cleanup_control_files(self):
        try:
            ctl_file = '%s.ctl' % self.pid_path
            LOG.debug('Removing %(pidfile)s and %(ctlfile)s',
                      {'pidfile': self.pid_file,
                       'ctlfile': ctl_file})

            if os.path.exists(self.pid_file):
                os.remove(self.pid_file)

            if os.path.exists(ctl_file):
                os.remove(ctl_file)

        except OSError as e:
            LOG.error('Unable to remove pluto control '
                      'files for router %(router)s. %(msg)s',
                      {'router': self.id, 'msg': e})

    def get_status(self):
        return self._execute([self.binary,
                              'whack',
                              '--ctlbase',
                              self.pid_path,
                              '--status'], extra_ok_codes=[1, 3])

    def _config_changed(self):
        secrets_file = os.path.join(
            self.etc_dir, 'ipsec.secrets')
        config_file = os.path.join(
            self.etc_dir, 'ipsec.conf')

        if not os.path.isfile(secrets_file + '.old'):
            return True
        if not os.path.isfile(config_file + '.old'):
            return True

        if not filecmp.cmp(secrets_file, secrets_file + '.old'):
            return True
        if not filecmp.cmp(config_file, config_file + '.old'):
            return True

        return False

    def restart(self):
        """Restart the process."""
        if cfg.CONF.pluto.restart_check_config and not self._config_changed():
            return
        # stop() followed immediately by a start() runs the risk that the
        # current pluto daemon has not had a chance to shutdown. We check
        # the current process information to see if the daemon is still
        # running and if so, wait a short interval and retry.
        self.stop()
        wait_interval = cfg.CONF.pluto.shutdown_check_timeout
        for i in range(cfg.CONF.pluto.shutdown_check_retries):
            if not self._process_running():
                self._cleanup_control_files()
                break
            eventlet.sleep(wait_interval)
            wait_interval *= cfg.CONF.pluto.shutdown_check_back_off
        else:
            LOG.warning('Server appears to still be running, restart '
                        'of router %s may fail', self.id)
        self.start()
        return

    def _resolve_fqdn(self, fqdn):
        # The first addrinfo member from the list returned by
        # socket.getaddrinfo is used for the address resolution.
        # The code doesn't filter for ipv4 or ipv6 address.
        try:
            addrinfo = socket.getaddrinfo(fqdn, None)[0]
            return addrinfo[-1][0]
        except socket.gaierror:
            LOG.exception("Peer address %s cannot be resolved", fqdn)

    def _get_nexthop(self, address, connection_id):
        # check if address is an ip address or fqdn
        invalid_ip_address = validators.validate_ip_address(address)
        if invalid_ip_address:
            ip_addr = self._resolve_fqdn(address)
            if not ip_addr:
                self._record_connection_status(connection_id, constants.ERROR,
                                               force_status_update=True)
                raise vpn_exception.VPNPeerAddressNotResolved(
                    peer_address=address)
        else:
            ip_addr = address
        routes = self._execute(['ip', 'route', 'get', ip_addr])
        if routes.find('via') >= 0:
            return routes.split(' ')[2]
        return address

    def _virtual_privates(self, vpnservice):
        """Returns line of virtual_privates.

        virtual_private contains the networks
        that are allowed as subnet for the remote client.
        """
        virtual_privates = []
        nets = []
        for ipsec_site_conn in vpnservice['ipsec_site_connections']:
            nets += ipsec_site_conn['local_cidrs']
            nets += ipsec_site_conn['peer_cidrs']
        for net in nets:
            version = netaddr.IPNetwork(net).version
            virtual_privates.append('%v{}:{}'.format(version, net))
        virtual_privates.sort()
        return ','.join(virtual_privates)

    def _gen_config_content(self, template_file, vpnservice):
        template = _get_template(template_file)
        virtual_privates = self._virtual_privates(vpnservice)
        return template.render(
            {'vpnservice': vpnservice,
             'virtual_privates': virtual_privates,
             'nat_traversal': self.IPSEC_CONF_NAT_TRAVERSAL})

    def start_pluto(self):
        cmd = [self.binary,
               'pluto',
               '--ctlbase', self.pid_path,
               '--ipsecdir', self.etc_dir,
               '--use-netkey',
               '--uniqueids',
               '--nat_traversal',
               '--secretsfile', self.secrets_file]

        if self.conf.ipsec.enable_detailed_logging:
            cmd += ['--perpeerlog', '--perpeerlogbase', self.log_dir]
        self._execute(cmd)

    def add_ipsec_connection(self, nexthop, conn_id):
        self._execute([self.binary,
                       'addconn',
                       '--ctlbase', '%s.ctl' % self.pid_path,
                       '--defaultroutenexthop', nexthop,
                       '--config', self.config_file, conn_id
                       ])

    def start_whack_listening(self):
        #TODO(nati) fix this when openswan is fixed
        #Due to openswan bug, this command always exit with 3
        self._execute([self.binary,
                       'whack',
                       '--ctlbase', self.pid_path,
                       '--listen'
                       ], check_exit_code=False)

    def shutdown_whack(self):
        self._execute([self.binary,
                       'whack',
                       '--ctlbase', self.pid_path,
                       '--shutdown'
                       ])

    def initiate_connection(self, conn_name):
        self._execute([self.binary,
                       'whack',
                       '--ctlbase', self.pid_path,
                       '--name', conn_name,
                       '--asynchronous',
                       '--initiate'
                       ])

    def terminate_connection(self, conn_name):
        self._execute([self.binary,
                       'whack',
                       '--ctlbase', self.pid_path,
                       '--name', conn_name,
                       '--terminate'
                       ])

    def start(self):
        """Start the process.

        Note: if there is not namespace yet,
        just do nothing, and wait next event.
        """
        if not self.namespace:
            return

        # NOTE: The restart operation calls the parent's start() instead of
        # this one to avoid having to special case the startup file check.
        # If anything is added to this method that needs to run whenever
        # a restart occurs, it should be either added to the restart()
        # override or things refactored to special-case start() when
        # called from restart().

        # If, by any reason, ctl and pid  files weren't cleaned up, pluto
        # won't be able to rewrite them and will fail to start. So we check
        # to see if the process is running and if not, attempt a cleanup.
        # In either case we fall through to allow the pluto process to
        # start or fail in the usual way.
        if not self._process_running():
            self._cleanup_control_files()

        #start pluto IKE keying daemon
        self.start_pluto()

        #add connections
        for ipsec_site_conn in self.vpnservice['ipsec_site_connections']:
            # Don't add a connection if its admin state is down
            if not ipsec_site_conn['admin_state_up']:
                continue
            nexthop = self._get_nexthop(ipsec_site_conn['peer_address'],
                                        ipsec_site_conn['id'])
            self.add_ipsec_connection(nexthop, ipsec_site_conn['id'])

        #start whack ipsec keying daemon
        self.start_whack_listening()

        for ipsec_site_conn in self.vpnservice['ipsec_site_connections']:
            if (not ipsec_site_conn['initiator'] == 'start' or
                    not ipsec_site_conn['admin_state_up']):
                continue
            #initiate ipsec connection
            self.initiate_connection(ipsec_site_conn['id'])
        self._copy_configs()

    def get_established_connections(self):
        connections = []
        status_output = self.get_status()

        if not status_output:
            return connections

        for line in status_output.split('\n'):
            if self.STATUS_NOT_RUNNING_PATTERN.search(line):
                return connections
            m = self.STATUS_IPSEC_SA_ESTABLISHED_PATTERN2.search(line)
            if m:
                connection = m.group(1)
                if connection in connections:
                    continue
                connections.append(connection)
        return connections

    def disconnect(self):
        if not self.namespace:
            return
        if not self.vpnservice:
            return

        connections = self.get_established_connections()
        for conn_name in connections:
            self.terminate_connection(conn_name)

    def stop(self):
        #Stop process using whack
        #Note this will also stop pluto
        self.disconnect()
        self.shutdown_whack()
        self.connection_status = {}


class IPsecVpnDriverApi:
    """IPSecVpnDriver RPC api."""

    @log_helpers.log_method_call
    def __init__(self, topic):
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    @log_helpers.log_method_call
    def get_vpn_services_on_host(self, context, host):
        """Get list of vpnservices.

        The vpnservices including related ipsec_site_connection,
        ikepolicy and ipsecpolicy on this host
        """
        cctxt = self.client.prepare()
        return cctxt.call(context, 'get_vpn_services_on_host', host=host)

    @log_helpers.log_method_call
    def update_status(self, context, status):
        """Update local status.

        This method call updates status attribute of
        VPNServices.
        """
        cctxt = self.client.prepare()
        return cctxt.call(context, 'update_status', status=status)


class IPsecDriver(device_drivers.DeviceDriver, metaclass=abc.ABCMeta):
    """VPN Device Driver for IPSec.

    This class is designed for use with L3-agent now.
    However this driver will be used with another agent in future
    so the use of "Router" is kept minimal now.
    Instead of router_id, we are using process_id in this code.
    """

    # history
    #   1.0 Initial version
    target = oslo_messaging.Target(version='1.0')

    def __init__(self, vpn_service, host):
        # TODO(pc_m) Replace vpn_service with config arg, once all driver
        # implementations no longer need vpn_service.
        self.conf = vpn_service.conf
        self.host = host
        self.conn = n_rpc.Connection()
        self.context = context.get_admin_context_without_session()
        self.topic = topics.IPSEC_AGENT_TOPIC
        node_topic = '{}.{}'.format(self.topic, self.host)

        self.processes = {}
        self.routers: ty.Dict[str, ty.Any] = {}
        self.process_status_cache = {}

        self.endpoints = [self]
        self.conn.create_consumer(node_topic, self.endpoints, fanout=False)
        self.conn.consume_in_threads()
        self.agent_rpc = IPsecVpnDriverApi(topics.IPSEC_DRIVER_TOPIC)
        self.process_status_cache_check = loopingcall.FixedIntervalLoopingCall(
            self.report_status, self.context)
        self.process_status_cache_check.start(
            interval=self.conf.ipsec.ipsec_status_check_interval)

    def get_namespace(self, router_id):
        """Get namespace of router.

        :router_id: router_id
        :returns: namespace string.
            Note: If the router is a DVR, then the SNAT namespace will be
                  provided. If the router does not exist, return None.
        """
        router_info = self.routers.get(router_id)
        if not router_info:
            return
        # For DVR, use SNAT namespace
        # TODO(pcm): Use router object method to tell if DVR, when available
        if router_info.router['distributed']:
            return router_info.snat_namespace.name
        return router_info.ns_name

    def get_router_based_iptables_manager(self, ri):
        """Returns router based iptables manager

        In DVR routers the IPsec VPN service should run inside
        the snat namespace. So the iptables manager used for
        snat namespace is different from the iptables manager
        used for the qr namespace in a non dvr based router.

        This function will check the router type and then will
        return the right iptables manager. If DVR enabled router
        it will return the snat_iptables_manager otherwise it will
        return the legacy iptables_manager.
        """
        # TODO(pcm): Use router object method to tell if DVR, when available
        if ri.router['distributed']:
            return ri.snat_iptables_manager
        return ri.iptables_manager

    def ensure_nat_rules(self, vpnservice):
        """Ensure the required nat rules for ipsec exist in iptables.

        :param vpnservice: vpnservices
        """
        LOG.debug("ensure_nat_rules called for router %s",
                  vpnservice['router_id'])
        ri = self.routers.get(vpnservice['router_id'])
        if not ri:
            LOG.debug("No router info for router %s", vpnservice['router_id'])
            return

        iptables_manager = self.get_router_based_iptables_manager(ri)
        # clear all existing rules first
        LOG.debug("Clearing vpnaas tagged NAT rules for router %s",
                  ri.router_id)
        iptables_manager.ipv4['nat'].clear_rules_by_tag('vpnaas')

        for ipsec_site_connection in vpnservice['ipsec_site_connections']:
            for local_cidr in ipsec_site_connection['local_cidrs']:
                # This ipsec rule is not needed for ipv6.
                if netaddr.IPNetwork(local_cidr).version == 6:
                    continue

                for peer_cidr in ipsec_site_connection['peer_cidrs']:
                    LOG.debug("Adding an ipsec policy NAT rule"
                              "%s <-> %s to router id %s",
                              peer_cidr, local_cidr, vpnservice['router_id'])

                    iptables_manager.ipv4['nat'].add_rule(
                        'POSTROUTING',
                        '-s %s -d %s -m policy '
                        '--dir out --pol ipsec '
                        '-j ACCEPT ' % (local_cidr, peer_cidr),
                        top=True, tag='vpnaas')

        LOG.debug("Applying iptables for router id %s",
                  vpnservice['router_id'])
        iptables_manager.apply()

    @log_helpers.log_method_call
    def remove_nat_rules(self, router_id):
        """Remove all our iptables rules in namespace.

        :param router_id: router_id
        """
        router = self.routers.get(router_id)
        if not router:
            return
        iptables_manager = self.get_router_based_iptables_manager(router)
        iptables_manager.ipv4['nat'].clear_rules_by_tag('vpnaas')
        iptables_manager.apply()

    @log_helpers.log_method_call
    def vpnservice_updated(self, context, **kwargs):
        """Vpnservice updated rpc handler

        VPN Service Driver will call this method
        when vpnservices updated.
        Then this method start sync with server.
        """
        router = kwargs.get('router', None)
        self.sync(context, [router] if router else [])

    @abc.abstractmethod
    def create_process(self, process_id, vpnservice, namespace):
        pass

    def ensure_process(self, process_id, vpnservice=None):
        """Ensuring process.

        If the process doesn't exist, it will create process
        and store it in self.process
        """
        process = self.processes.get(process_id)
        if not process or not process.namespace:
            namespace = self.get_namespace(process_id)
            process = self.create_process(
                process_id,
                vpnservice,
                namespace)
            self.processes[process_id] = process
        elif vpnservice:
            process.update_vpnservice(vpnservice)
        return process

    def create_router(self, router_info: RouterInfo):
        """Handling create router event.

        Agent calls this method, when the process namespace is ready.
        Note: process_id == router_id == vpnservice_id
        """
        process_id = router_info.router_id
        self.routers[process_id] = router_info
        if process_id in self.processes:
            # In case of vpnservice is created
            # before router's namespace
            process = self.processes[process_id]
            self.ensure_nat_rules(process.vpnservice)
            # Don't run ipsec process for backup HA router
            if router_info.router['ha'] and router_info.ha_state == 'backup':
                return
            process.enable()

    def destroy_process(self, process_id):
        """Destroy process.

        Disable the process, remove the iptables rules, and remove the process
        manager for the processes that no longer are running vpn service.
        """
        if process_id in self.processes:
            process = self.processes[process_id]
            process.disable()
            vpnservice = process.vpnservice
            if vpnservice:
                self.remove_nat_rules(process_id)
            del self.processes[process_id]

    def destroy_router(self, process_id):
        """Handling destroy_router event.

        Agent calls this method, when the process namespace
        is deleted.
        """
        self.destroy_process(process_id)
        if process_id in self.routers:
            del self.routers[process_id]

    def get_process_status_cache(self, process):
        if not self.process_status_cache.get(process.id):
            self.process_status_cache[process.id] = {
                'status': None,
                'id': process.vpnservice['id'],
                'updated_pending_status': False,
                'ipsec_site_connections': {}}
        return self.process_status_cache[process.id]

    def is_status_updated(self, process, previous_status):
        if process.updated_pending_status:
            return True
        if process.status != previous_status['status']:
            return True
        if (process.connection_status !=
            previous_status['ipsec_site_connections']):
            return True

    def unset_updated_pending_status(self, process):
        process.updated_pending_status = False
        for connection_status in process.connection_status.values():
            connection_status['updated_pending_status'] = False

    def copy_process_status(self, process):
        return {
            'id': process.vpnservice['id'],
            'status': process.status,
            'updated_pending_status': process.updated_pending_status,
            'ipsec_site_connections': copy.deepcopy(process.connection_status)
        }

    def update_downed_connections(self, process_id, new_status):
        """Update info to be reported, if connections just went down.

        If there is no longer any information for a connection, because it
        has been removed (e.g. due to an admin down of VPN service or IPSec
        connection), but there was previous status information for the
        connection, mark the connection as down for reporting purposes.
        """
        if process_id in self.process_status_cache:
            for conn in self.process_status_cache[process_id][IPSEC_CONNS]:
                if conn not in new_status[IPSEC_CONNS]:
                    new_status[IPSEC_CONNS][conn] = {
                        'status': constants.DOWN,
                        'updated_pending_status': True
                    }

    def should_be_reported(self, context, process):
        if (context.is_admin or
            process.vpnservice["tenant_id"] == context.tenant_id):
            return True

    @log_helpers.log_method_call
    def report_status(self, context):
        status_changed_vpn_services = []
        for process_id, process in list(self.processes.items()):
            # NOTE(mnaser): It's not necessary to check status for processes
            #               of a backup L3 agent
            router = self.routers.get(process_id)
            if router and router.router['ha'] and router.ha_state == 'backup':
                LOG.debug("%s router in backup state, skipping", process_id)
                continue
            if not self.should_be_reported(context, process):
                continue
            previous_status = self.get_process_status_cache(process)
            if self.is_status_updated(process, previous_status):
                new_status = self.copy_process_status(process)
                self.update_downed_connections(process.id, new_status)
                status_changed_vpn_services.append(new_status)
                self.process_status_cache[process.id] = (
                    self.copy_process_status(process))
                # We need unset updated_pending status after it
                # is reported to the server side
                self.unset_updated_pending_status(process)

        if status_changed_vpn_services:
            self.agent_rpc.update_status(
                context,
                status_changed_vpn_services)

    @log_helpers.log_method_call
    @lockutils.synchronized('vpn-agent', 'neutron-')
    def sync(self, context, router_information: ty.List):
        """Sync status with server side.

        :param context: context object for RPC call
        :param router_information: RouterInfo objects with updated state

        There could be many failure cases should be
        considered including the followings.
        1) Agent class restarted
        2) Failure on process creation
        3) VpnService is deleted during agent down
        4) RPC failure

        In order to handle these failure cases,
        This driver takes simple sync strategies.
        """
        vpnservices = self.agent_rpc.get_vpn_services_on_host(
            context, self.host)
        router_ids = [vpnservice['router_id'] for vpnservice in vpnservices]
        sync_router_ids = []

        for ri in router_information:
            # Update our router_info with the updated one
            if isinstance(ri, RouterInfo):
                router_id = ri.router_id if ri.router_id else None
            elif isinstance(ri, dict):
                router_id = ri.get("router_id")
            else:
                router_id = None

            if not router_id:
                continue

            self.routers[router_id] = ri
            sync_router_ids.append(router_id)

        self._sync_vpn_processes(vpnservices, sync_router_ids)
        self._delete_vpn_processes(sync_router_ids, router_ids)
        self._cleanup_stale_vpn_processes(router_ids)

        self.report_status(context)

    def _sync_vpn_processes(self, vpnservices, sync_router_ids):
        # Ensure the ipsec process is enabled only for
        # - the vpn services which are not yet in self.processes
        # - vpn services whose router id is in 'sync_router_ids'
        for vpnservice in vpnservices:
            if vpnservice['router_id'] not in self.processes or (
                    vpnservice['router_id'] in sync_router_ids):
                process = self.ensure_process(vpnservice['router_id'],
                                              vpnservice=vpnservice)
                self.ensure_nat_rules(vpnservice)
                ri = self.routers.get(vpnservice['router_id'])
                if not ri:
                    continue
                # For HA router, spawn vpn process on master router
                # and terminate vpn process on backup router
                if ri.router['ha'] and ri.ha_state == 'backup':
                    process.disable()
                else:
                    process.update()

    def _delete_vpn_processes(self, sync_router_ids, vpn_router_ids):
        # Delete any IPSec processes that are
        # associated with routers, but are not running the VPN service.
        for process_id in sync_router_ids:
            if process_id not in vpn_router_ids:
                self.destroy_process(process_id)

    def _cleanup_stale_vpn_processes(self, vpn_router_ids):
        # Delete any IPSec processes running
        # VPN that do not have an associated router.
        process_ids = [pid for pid in self.processes
                       if pid not in vpn_router_ids]
        for process_id in process_ids:
            self.destroy_process(process_id)


class OpenSwanDriver(IPsecDriver):
    def create_process(self, process_id, vpnservice, namespace):
        return OpenSwanProcess(
            self.conf,
            process_id,
            vpnservice,
            namespace)



#    (c) Copyright 2015 Cisco Systems Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""VPNaaS endpoint groups

Revision ID: 41b509d10b5e
Revises: 24f28869838b
Create Date: 2015-08-06 18:21:03.241664

"""

from alembic import op
import sqlalchemy as sa

from neutron_vpnaas.services.vpn.common import constants

# revision identifiers, used by Alembic.
revision = '41b509d10b5e'
down_revision = '24f28869838b'


def upgrade():
    op.create_table(
        'vpn_endpoint_groups',
        sa.Column('id', sa.String(length=36), nullable=False,
                  primary_key=True),
        sa.Column('tenant_id', sa.String(length=255),
                  index=True),
        sa.Column('name', sa.String(length=255)),
        sa.Column('description', sa.String(length=255)),
        sa.Column('endpoint_type',
                  sa.Enum(constants.SUBNET_ENDPOINT, constants.CIDR_ENDPOINT,
                          constants.VLAN_ENDPOINT, constants.NETWORK_ENDPOINT,
                          constants.ROUTER_ENDPOINT,
                          name='endpoint_type'),
                  nullable=False),
    )
    op.create_table(
        'vpn_endpoints',
        sa.Column('endpoint', sa.String(length=255), nullable=False),
        sa.Column('endpoint_group_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['endpoint_group_id'],
                                ['vpn_endpoint_groups.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('endpoint', 'endpoint_group_id'),
    )



# Copyright (c) 2013 OpenStack Foundation.
# Copyright (c) 2023 SysEleven GmbH.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from neutron.extensions import router_availability_zone as router_az
from neutron import worker as neutron_worker
from neutron_lib import context as ncontext
from neutron_lib.db import api as db_api
from neutron_lib.db import model_base
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
import oslo_messaging
import sqlalchemy as sa
from sqlalchemy import func

from neutron_vpnaas._i18n import _
from neutron_vpnaas.db.vpn import vpn_models
from neutron_vpnaas.extensions import vpn_agentschedulers
from neutron_vpnaas.services.vpn.common.constants import AGENT_TYPE_VPN


LOG = logging.getLogger(__name__)

VPN_AGENTS_SCHEDULER_OPTS = [
    cfg.StrOpt('vpn_scheduler_driver',
               default='neutron_vpnaas.scheduler.vpn_agent_scheduler'
                       '.LeastRoutersScheduler',
               help=_('Driver to use for scheduling '
                      'router to a VPN agent')),
    cfg.BoolOpt('vpn_auto_schedule', default=True,
                help=_('Allow auto scheduling of routers to VPN agent.')),
    cfg.BoolOpt('allow_automatic_vpnagent_failover', default=False,
                help=_('Automatically reschedule routers from offline VPN '
                       'agents to online VPN agents.')),
]

cfg.CONF.register_opts(VPN_AGENTS_SCHEDULER_OPTS)


class RouterVPNAgentBinding(model_base.BASEV2):
    """Represents binding between neutron routers and VPN agents."""

    router_id = sa.Column(sa.String(36),
                          sa.ForeignKey("routers.id", ondelete='CASCADE'),
                          primary_key=True,
                          unique=True,
                          nullable=False)
    vpn_agent_id = sa.Column(sa.String(36), primary_key=True, nullable=False)


class VPNAgentSchedulerDbMixin(
    vpn_agentschedulers.VPNAgentSchedulerPluginBase):
    """Mixin class to add VPN agent scheduler extension to plugins
    using the VPN agent.
    """

    vpn_scheduler = None
    agent_notifiers = {}

    @property
    def l3_plugin(self):
        return directory.get_plugin(plugin_const.L3)

    @property
    def core_plugin(self):
        return directory.get_plugin()

    def add_periodic_vpn_agent_status_check(self):
        if not cfg.CONF.allow_automatic_vpnagent_failover:
            LOG.info("Skipping periodic VPN agent status check because "
                     "automatic rescheduling is disabled.")
            return

        interval = max(cfg.CONF.agent_down_time // 2, 1)
        # add random initial delay to allow agents to check in after the
        # neutron server first starts. random to offset multiple servers
        initial_delay = random.randint(interval, interval * 2)

        check_worker = neutron_worker.PeriodicWorker(
            self.reschedule_vpnservices_from_down_agents,
            interval, initial_delay)
        self.add_worker(check_worker)

    def reschedule_vpnservices_from_down_agents(self):
        """Reschedule VPN services from down VPN agents.

        VPN services are scheduled per router.
        """
        context = ncontext.get_admin_context()
        try:
            down_bindings = self.get_down_router_bindings(context)

            agents_back_online = set()
            for binding in down_bindings:
                if binding.vpn_agent_id in agents_back_online:
                    continue
                agent = self.core_plugin.get_agent(context,
                                                   binding.vpn_agent_id)
                if agent['alive']:
                    agents_back_online.add(binding.vpn_agent_id)
                    continue

                LOG.warning(
                    "Rescheduling vpn services for router %(router)s from "
                    "agent %(agent)s because the agent is not alive.",
                    {'router': binding.router_id,
                     'agent': binding.vpn_agent_id})
                try:
                    self.reschedule_router(context, binding.router_id, agent)
                except (vpn_agentschedulers.RouterReschedulingFailed,
                        oslo_messaging.RemoteError):
                    # Catch individual rescheduling errors here
                    # so one broken one doesn't stop the iteration.
                    LOG.exception("Failed to reschedule vpn services for "
                                  "router %s", binding.router_id)
        except Exception:
            # we want to be thorough and catch whatever is raised
            # to avoid loop abortion
            LOG.exception("Exception encountered during vpn service "
                          "rescheduling.")

    @db_api.CONTEXT_READER
    def get_down_router_bindings(self, context):
        vpn_agents = self.get_vpn_agents(context, active=False)
        if not vpn_agents:
            return []
        vpn_agent_ids = [vpn_agent['id'] for vpn_agent in vpn_agents]

        query = context.session.query(RouterVPNAgentBinding)
        query = query.filter(
            RouterVPNAgentBinding.vpn_agent_id.in_(vpn_agent_ids))
        return query.all()

    def validate_agent_router_combination(self, context, agent, router):
        """Validate if the router can be correctly assigned to the agent.

        :raises: InvalidVPNAgent if attempting to assign router to an
          unsuitable agent (disabled, type != VPN, incompatible configuration)
        """
        if agent['agent_type'] != AGENT_TYPE_VPN:
            raise vpn_agentschedulers.InvalidVPNAgent(id=agent['id'])

    @db_api.CONTEXT_READER
    def check_agent_router_scheduling_needed(self, context, agent, router):
        """Check if the scheduling of router's VPN services is needed.

        :raises: RouterHostedByVPNAgent if router is already assigned
          to a different agent.
        :returns: True if scheduling is needed, otherwise False
        """
        router_id = router['id']
        agent_id = agent['id']
        query = context.session.query(RouterVPNAgentBinding)
        bindings = query.filter_by(router_id=router_id).all()
        if not bindings:
            return True
        for binding in bindings:
            if binding.vpn_agent_id == agent_id:
                # router already bound to the agent we need
                return False
        # Router is already bound to some agent
        raise vpn_agentschedulers.RouterHostedByVPNAgent(
            router_id=router_id,
            agent_id=bindings[0].vpn_agent_id)

    def create_router_to_agent_binding(self, context, router_id, agent_id):
        """Create router to VPN agent binding."""
        try:
            with db_api.CONTEXT_WRITER.using(context):
                binding = RouterVPNAgentBinding()
                binding.vpn_agent_id = agent_id
                binding.router_id = router_id
                context.session.add(binding)
        except db_exc.DBDuplicateEntry:
            LOG.debug('VPN service of router %(router_id)s has already been '
                      'scheduled to a VPN agent.',
                      {'router_id': router_id})
            return False
        except db_exc.DBReferenceError:
            LOG.debug('Router %s has already been removed '
                      'by concurrent operation', router_id)
            return False

        LOG.debug('VPN service of router %(router_id)s is scheduled to '
                  'VPN agent %(agent_id)s',
                  {'router_id': router_id, 'agent_id': agent_id})
        return True

    def add_router_to_vpn_agent(self, context, agent_id, router_id):
        """Add a VPN agent to host VPN services of a router."""
        with db_api.CONTEXT_WRITER.using(context):
            router = self.l3_plugin.get_router(context, router_id)
            agent = self.core_plugin.get_agent(context, agent_id)
            self.validate_agent_router_combination(context, agent, router)
            if not self.check_agent_router_scheduling_needed(
                    context, agent, router):
                return
        try:
            success = self.create_router_to_agent_binding(
                context, router['id'], agent['id'])
        except db_exc.DBError:
            success = False

        if not success:
            raise vpn_agentschedulers.RouterSchedulingFailed(
                router_id=router_id, agent_id=agent_id)

        # notify agent
        vpn_notifier = self.agent_notifiers.get(AGENT_TYPE_VPN)
        if vpn_notifier:
            vpn_notifier.vpnservice_added_to_agent(
                context, [router_id], agent['host'])

        # update port binding
        self.vpn_router_agent_binding_changed(
            context, router_id, agent['host'])

    def remove_router_from_vpn_agent(self, context, agent_id, router_id):
        """Remove the router from VPN agent.

        After removal, the VPN service(s) of the router will be non-hosted
        until there is an update which leads to re-schedule or the router is
        added to another agent manually.
        """
        agent = self.core_plugin.get_agent(context, agent_id)

        self._unbind_router(context, router_id, agent_id)

        vpn_notifier = self.agent_notifiers.get(AGENT_TYPE_VPN)
        if vpn_notifier:
            vpn_notifier.vpnservice_removed_from_agent(
                context, router_id, agent['host'])

    def _unbind_router(self, context, router_id, agent_id):
        with db_api.CONTEXT_WRITER.using(context):
            query = context.session.query(RouterVPNAgentBinding)
            query = query.filter(
                RouterVPNAgentBinding.router_id == router_id,
                RouterVPNAgentBinding.vpn_agent_id == agent_id)
            return query.delete()

    def reschedule_router(self, context, router_id, cur_agent):
        """Reschedule router to a new VPN agent

        Remove the router from the agent currently hosting it and
        schedule it again
        """
        with db_api.CONTEXT_WRITER.using(context):
            deleted = self._unbind_router(context, router_id, cur_agent['id'])
            if not deleted:
                # If nothing was deleted, the binding didn't exist anymore
                # because some other server deleted the binding concurrently.
                # Stop here.
                return

            new_agent = self.schedule_router(context, router_id)
            if not new_agent:
                # No new_agent means that another server scheduled the
                # router concurrently. Don't raise RouterReschedulingFailed.
                return

        self._notify_agents_router_rescheduled(context, router_id,
                                               cur_agent, new_agent)
        # update port binding
        self.vpn_router_agent_binding_changed(
            context, router_id, new_agent['host'])

    def _notify_agents_router_rescheduled(self, context, router_id,
                                          old_agent, new_agent):
        vpn_notifier = self.agent_notifiers.get(AGENT_TYPE_VPN)
        if not vpn_notifier:
            return

        old_host = old_agent['host']
        new_host = new_agent['host']
        if old_host != new_host:
            vpn_notifier.vpnservice_removed_from_agent(
                context, router_id, old_host)

            try:
                vpn_notifier.vpnservice_added_to_agent(
                    context, [router_id], new_host)
            except oslo_messaging.MessagingException:
                self._unbind_router(context, router_id, new_agent['id'])
                raise vpn_agentschedulers.RouterReschedulingFailed(
                    router_id=router_id)

    @db_api.CONTEXT_READER
    def list_routers_on_vpn_agent(self, context, agent_id):
        query = context.session.query(RouterVPNAgentBinding.router_id)
        query = query.filter(RouterVPNAgentBinding.vpn_agent_id == agent_id)

        router_ids = [item[0] for item in query]
        if router_ids:
            return {'routers':
                    self.l3_plugin.get_routers(context,
                                               filters={'id': router_ids})}
        else:
            # Exception will be thrown if the requested agent does not exist.
            self.core_plugin.get_agent(context, agent_id)
            return {'routers': []}

    @db_api.CONTEXT_READER
    def get_vpn_agents_hosting_routers(self, context, router_ids, active=None):
        if not router_ids:
            return []
        query = context.session.query(RouterVPNAgentBinding)
        query = query.filter(RouterVPNAgentBinding.router_id.in_(router_ids))

        filters = {'id': [binding.vpn_agent_id for binding in query]}
        vpn_agents = self.core_plugin.get_agents(context, filters=filters)
        if active is not None:
            vpn_agents = [agent
                          for agent in vpn_agents
                          if agent['alive'] == active]
        return vpn_agents

    def list_vpn_agents_hosting_router(self, context, router_id):
        vpn_agents = self.get_vpn_agents_hosting_routers(context, [router_id])
        return {'agents': vpn_agents}

    def get_vpn_agents(self, context, active=None, host=None):
        filters = {'agent_type': [AGENT_TYPE_VPN]}
        if host is not None:
            filters['host'] = [host]
        vpn_agents = self.core_plugin.get_agents(context, filters=filters)
        if active is None:
            return vpn_agents
        else:
            return [vpn_agent
                    for vpn_agent in vpn_agents
                    if vpn_agent['alive'] == active]

    def get_vpn_agent_on_host(self, context, host, active=None):
        agents = self.get_vpn_agents(context, active=active, host=host)
        if agents:
            return agents[0]

    @db_api.CONTEXT_READER
    def get_unscheduled_vpn_routers(self, context, router_ids=None):
        """Get IDs of routers which have unscheduled VPN services."""
        query = context.session.query(vpn_models.VPNService.router_id)
        query = query.outerjoin(
            RouterVPNAgentBinding,
            vpn_models.VPNService.router_id == RouterVPNAgentBinding.router_id)
        query = query.filter(RouterVPNAgentBinding.vpn_agent_id.is_(None))
        if router_ids:
            query = query.filter(
                vpn_models.VPNService.router_id.in_(router_ids))
        return [router_id for router_id, in query.all()]

    def auto_schedule_routers(self, context, vpn_agent):
        if self.vpn_scheduler:
            return self.vpn_scheduler.auto_schedule_routers(
                self, context, vpn_agent)

    def schedule_router(self, context, router, candidates=None):
        """Schedule VPN services of a router to a VPN agent.

        Returns the chosen agent; None if another server scheduled the
        router concurrently.
        Raises RouterReschedulingFailed if no suitable agent is found.
        """
        if self.vpn_scheduler:
            return self.vpn_scheduler.schedule(
                self, context, router, candidates=candidates)

    @db_api.CONTEXT_READER
    def get_vpn_agent_with_min_routers(self, context, agent_ids):
        """Return VPN agent with the least number of routers."""
        if not agent_ids:
            return None
        query = context.session.query(
            RouterVPNAgentBinding.vpn_agent_id,
            func.count(RouterVPNAgentBinding.router_id).label('count'))
        query = query.group_by(RouterVPNAgentBinding.vpn_agent_id)
        query = query.order_by('count')
        query = query.filter(RouterVPNAgentBinding.vpn_agent_id.in_(agent_ids))
        used_agent_ids = [agent_id for agent_id, _ in query.all()]
        unused_agent_ids = set(agent_ids) - set(used_agent_ids)
        if unused_agent_ids:
            return unused_agent_ids.pop()
        else:
            return used_agent_ids[0]

    def get_hosts_to_notify(self, context, router_id):
        """Returns all hosts to send notification about router update"""
        agents = self.get_vpn_agents_hosting_routers(context, [router_id],
                                                     active=True)
        return [a['host'] for a in agents]


class AZVPNAgentSchedulerDbMixin(VPNAgentSchedulerDbMixin,
                                 router_az.RouterAvailabilityZonePluginBase):
    """Mixin class to add availability_zone supported VPN agent scheduler."""

    def get_router_availability_zones(self, router):
        return list({agent.availability_zone for agent in router.vpn_agents})
